# -*- coding: utf-8 -*-
__author__ = 'ikibalin'
__version__ = "2020_03_05"

import os
import os.path
import sys

from PyQt5 import QtCore, QtGui, QtWidgets

from cryspy_editor.b_rcif_to_cryspy import rcif_to_cryspy
from cryspy_editor.widgets.FUNCTIONS import show_widget
from cryspy_editor.widgets.w_data_constr import w_for_data_constr
from cryspy_editor.widgets.w_global_constr import w_for_global_constr
from cryspy_editor.widgets.w_item_constr import w_for_item_constr
from cryspy_editor.widgets.w_loop_constr import w_for_loop_constr
from cryspy_editor.widgets.w_meas_inelasticl import w_for_meas_inelasticl
from cryspy_editor.widgets.w_pd2d_meas import w_for_pd2d_meas
from cryspy_editor.widgets.w_pd2d_proc import w_for_pd2d_proc
from cryspy_editor.widgets.w_pd_peak import w_for_pd_peak
from cryspy_editor.widgets.w_pd_meas import w_for_pd_meas
from cryspy_editor.widgets.w_pd_proc import w_for_pd_proc
from cryspy_editor.widgets.w_pd import w_for_pd
from cryspy_editor.widgets.w_diffrn_refln import w_for_diffrn_refln
from pycifstar import Data, Global, Item, Items, Loop, to_global

from cryspy_editor.widgets.w_item import w_for_item
from cryspy_editor.widgets.w_loop import w_for_loop

try:
    FLAG_CRYSPY = True
    from cryspy_editor.widgets.FUNCTIONS import del_layout
    from cryspy.common.cl_global_constr import GlobalConstr
    from cryspy.common.cl_global_constr import DataConstr
    from cryspy.common.cl_loop_constr import LoopConstr
    from cryspy.common.cl_item_constr import ItemConstr

    from cryspy.scripts.cl_rhochi import RhoChi
    from cryspy.cif_like.cl_crystal import Crystal
    from cryspy.cif_like.cl_pd import Pd
    from cryspy.cif_like.cl_pd2d import Pd2d
    from cryspy.cif_like.cl_diffrn import Diffrn
    from cryspy.cif_like.cl_diffrn_refln import DiffrnReflnL 

    from cryspy.pd1dcif_like.cl_pd_peak import PdPeakL
    from cryspy.pd1dcif_like.cl_pd_meas import PdMeasL
    from cryspy.pd1dcif_like.cl_pd_proc import PdProcL
    from cryspy.pd2dcif_like.cl_pd2d_meas import Pd2dMeas
    from cryspy.pd2dcif_like.cl_pd2d_proc import Pd2dProc
except ImportError:
    FLAG_CRYSPY = False

try:
    FLAG_MAGREF = True
    from magref import CrystRef
    from magref.classes.cl_cryst_field import CrystField
    from magref.classes.cl_meas_inelastic import MeasInelasticL
except ImportError:
    FLAG_MAGREF = False

try:
    FLAG_MEM = True
    from lib_mem.mem.cl_density import Density
    from lib_mem.scripts.cl_mem_tensor import MemTensor
except ImportError:
    FLAG_MEM = False






class MyThread(QtCore.QThread):
    signal_start = QtCore.pyqtSignal()
    signal_end = QtCore.pyqtSignal()
    signal_refresh = QtCore.pyqtSignal()
    signal_take_attributes = QtCore.pyqtSignal()

    def __init__(self, parent=None):
        QtCore.QThread.__init__(self, parent)
        self.message = None
        self.function = None
        self.arguments = None
        self.output = None

    def run(self):
        l_message = []
        self.signal_start.emit()
        w_cpanel = self.w_cpanel
        l_cpanel = w_cpanel.layout()

        func = self.function
        arg = self.arguments
        qtree_widget = None
        if l_cpanel.count() > 0:
            qtree_widget = l_cpanel.itemAt(0).widget()
        if isinstance(qtree_widget, QtWidgets.QTreeWidget):
            arg_new = []
            for _arg in arg:
                flag_arg = False
                if isinstance(_arg, str):
                    l_arg = _arg.split(",")
                    s_0 = l_arg[0]
                    level_item_count = qtree_widget.topLevelItemCount()
                    for _i in range(level_item_count):
                        qtree_widget_item = qtree_widget.topLevelItem(_i)
                        s_item = str(qtree_widget_item.text(0))
                        if (s_item.lower() == s_0.lower()):
                            arg_new.append(qtree_widget_item._object)
                            flag_arg = True
                            break
                if not flag_arg:
                    arg_new.append(_arg)
        else:
            arg_new = arg
        try:
            out = func(*arg_new)
        except:
            out = "Failed"
        self.output = out
        self.signal_end.emit()


class cbuilder(QtWidgets.QMainWindow):
    def __init__(self, f_dir_data=os.path.dirname(__file__)):
        super(cbuilder, self).__init__()

        self.__f_dir_prog = os.path.dirname(__file__)
        self.__f_dir_data = f_dir_data
        self.__f_file = None
        if os.path.isfile(f_dir_data):
            self.__f_file = f_dir_data
            self.__f_dir_data = os.path.dirname(f_dir_data)
        self.__mode = "star"

        self.def_actions()
        self.init_widget()
        self.setWindowTitle(f'CrysPy editor')

        if self.__f_file is not None:
            self.to_star()
        self.show()

    def def_actions(self):
        f_dir_prog = self.__f_dir_prog
        f_dir_prog_icon = os.path.join(f_dir_prog, 'f_icon')
        self.setWindowIcon(QtGui.QIcon(
            os.path.join(f_dir_prog_icon, 'icon.png')))

        open_action = QtWidgets.QAction(QtGui.QIcon(
            os.path.join(f_dir_prog_icon, 'open.png')), '&Open', self)
        open_action.setShortcut('Ctrl+O')
        open_action.setStatusTip('Open file')
        open_action.triggered.connect(self.open)

        to_star_action = QtWidgets.QAction(QtGui.QIcon(os.path.join(f_dir_prog_icon, 'to_star.png')), 'S&TAR mode',
                                           self)
        to_star_action.setStatusTip('Switch on STAR mode')
        to_star_action.triggered.connect(self.to_star)

        if FLAG_CRYSPY:
            to_cryspy_action = QtWidgets.QAction(QtGui.QIcon(os.path.join(f_dir_prog_icon, 'to_cryspy.png')),
                                                 '&CrysPy mode', self)
            to_cryspy_action.setStatusTip('Switch on CrysPy mode')
            to_cryspy_action.triggered.connect(self.to_cryspy)

            to_rhochi_action = QtWidgets.QAction(QtGui.QIcon(os.path.join(f_dir_prog_icon, 'to_rhochi.png')),
                                                 '&RhoChi mode', self)
            to_rhochi_action.setStatusTip('Switch on RhoChi mode')
            to_rhochi_action.triggered.connect(self.to_rhochi)

        if FLAG_MEM:
            to_mem_action = QtWidgets.QAction(QtGui.QIcon(os.path.join(f_dir_prog_icon, 'to_mem.png')), '&MEM mode',
                                              self)
            to_mem_action.setStatusTip('Switch on MEM mode')
            to_mem_action.triggered.connect(self.to_mem)

        if FLAG_MAGREF:
            to_magref_action = QtWidgets.QAction(QtGui.QIcon(os.path.join(f_dir_prog_icon, 'to_magref.png')),
                                                 'M&agRef mode', self)
            to_magref_action.setStatusTip('Switch on Magref mode')
            to_magref_action.triggered.connect(self.to_magref)

        save_action = QtWidgets.QAction(QtGui.QIcon(
            os.path.join(f_dir_prog_icon, 'save.png')), '&Save', self)
        save_action.setShortcut('Ctrl+S')
        save_action.setStatusTip('Save')
        save_action.triggered.connect(self.save)

        save_as_action = QtWidgets.QAction(QtGui.QIcon(os.path.join(f_dir_prog_icon, 'save_as.png')), 'Save &as...',
                                           self)
        save_as_action.setStatusTip('Save as ...')
        save_as_action.triggered.connect(self.save_as)

        open_folder = QtWidgets.QAction(QtGui.QIcon(os.path.join(f_dir_prog_icon, 'open_folder.png')), 'Open folder',
                                        self)
        open_folder.setStatusTip('Open folder')
        open_folder.triggered.connect(self.open_folder)

        self.statusBar()

        menubar = self.menuBar()
        fileMenu = menubar.addMenu('&File')
        fileMenu.addAction(open_action)
        fileMenu.addAction(to_star_action)
        if FLAG_CRYSPY:
            fileMenu.addAction(to_cryspy_action)
            fileMenu.addAction(to_rhochi_action)
        if FLAG_MEM:
            fileMenu.addAction(to_mem_action)
        if FLAG_MAGREF:
            fileMenu.addAction(to_magref_action)
        fileMenu.addAction(save_action)
        fileMenu.addAction(save_as_action)

        self.toolbar = self.addToolBar("Open")
        self.toolbar.addAction(open_action)
        self.toolbar.addAction(to_star_action)
        if FLAG_CRYSPY:
            self.toolbar.addAction(to_cryspy_action)
            self.toolbar.addAction(to_rhochi_action)
        if FLAG_MEM:
            self.toolbar.addAction(to_mem_action)
        if FLAG_MAGREF:
            self.toolbar.addAction(to_magref_action)
        self.toolbar.addAction(save_action)
        self.toolbar.addAction(save_as_action)
        self.toolbar.addAction(open_folder)

    def init_widget(self):
        """
module is reloaded for cwidget
        """
        self.location_on_the_screen()
        self.__widg_central = cwidget(self.info_width, self.info_height)
        self.setCentralWidget(self.__widg_central)

    def location_on_the_screen(self):
        """
        position and size of main window
        """
        screen = QtWidgets.QDesktopWidget().screenGeometry()
        self.setMinimumSize(screen.width() * 1 / 4, screen.height() * 1 / 4)
        self.info_width = screen.width() * 8. / 10.
        self.info_height = screen.height() * 14. / 16.
        self.move(screen.width() / 10, screen.height() / 20)
        self.resize(screen.width() * 8. / 10., screen.height() * 14. / 16.)

    def calc_is_finished(self):
        m_box = QtWidgets.QMessageBox()
        m_box.setIcon(QtWidgets.QMessageBox.Information)
        m_box.setText("Calculations are finished")
        m_box.setWindowTitle("Message")
        m_box.setStandardButtons(
            QtWidgets.QMessageBox.Ok | QtWidgets.QMessageBox.Cancel)
        m_box.exec()
        if self.__mode == "rhochi":
            self.to_rhochi()
        elif self.__mode == "mem":
            self.to_mem()
        elif self.__mode == "magref":
            self.to_magref()

    def to_star(self):
        if self.__widg_central._star is not None:
            star_ = self.__widg_central._star
        elif self.__f_file is not None:
            star_ = to_global(self.__f_file)
        else:
            f_file_data_new, okPressed = QtWidgets.QFileDialog.getOpenFileName(self, 'Select a cif file:',
                                                                               self.__f_dir_data, "All files (*.*)")
            if not (okPressed):
                return None
            star_ = to_global(f_file_data_new)
            self.__f_file = f_file_data_new
            self.__f_dir_data = os.path.dirname(f_file_data_new)
        self.__mode = "star"
        self.__widg_central.form_cpanel(star_)

    def to_cryspy(self):
        mode_orig = self.__mode
        flag = True
        cryspy_obj = None
        str_obj = None
        if mode_orig == "cryspy":
            cryspy_obj = self.__widg_central._object
        elif mode_orig == "star":
            obj = self.__widg_central._star
            if obj is not None:
                cryspy_obj = rcif_to_cryspy(obj)
        else:
            obj = self.__widg_central._object
            str_obj = str(obj.to_cif())
            star_obj = Global()
            star_obj.take_from_string(str_obj, f_dir=self.__f_dir_data)
            cryspy_obj = rcif_to_cryspy(star_obj)
        if cryspy_obj is None:
            cryspy_obj = GlobalConstr()
        if flag:
            self.__mode = "cryspy"
            cryspy_obj.file_input = self.__f_file
            self.__widg_central.form_cpanel(cryspy_obj)

    def to_rhochi(self):
        mode_orig = self.__mode
        flag = True
        cryspy_obj = None
        if mode_orig == "rhochi":
            cryspy_obj = self.__widg_central._object
        elif mode_orig == "star":
            obj = self.__widg_central._star
            if obj is not None:
                cryspy_obj = RhoChi.from_cif(str(obj))
        else:
            obj = self.__widg_central._object
            cryspy_obj = RhoChi.from_cif(obj.to_cif())
            if cryspy_obj is None:
                if mode_orig != "cryspy":
                    flag = False
        if cryspy_obj is None:
            cryspy_obj = RhoChi()
        if flag:
            self.__mode = "rhochi"
            cryspy_obj.file_input = self.__f_file
            self.__widg_central.form_cpanel(cryspy_obj)

    def to_mem(self):
        mode_orig = self.__mode
        flag = True
        cryspy_obj = None
        if mode_orig == "mem":
            cryspy_obj = self.__widg_central._object
        elif mode_orig == "star":
            obj = self.__widg_central._star
            if obj is not None:
                cryspy_obj = MemTensor.from_cif(str(obj))
        else:
            obj = self.__widg_central._object
            cryspy_obj = MemTensor.from_cif(obj.to_cif())
            if cryspy_obj is None:
                if mode_orig != "cryspy":
                    flag = False
        if cryspy_obj is None:
            cryspy_obj = MemTensor()
        if flag:
            self.__mode = "mem"
            cryspy_obj.file_input = self.__f_file
            self.__widg_central.form_cpanel(cryspy_obj)

    def to_magref(self):
        mode_orig = self.__mode
        flag = True
        cryspy_obj = None
        if mode_orig == "magref":
            cryspy_obj = self.__widg_central._object
        elif mode_orig == "star":
            obj = self.__widg_central._star
            if obj is not None:
                cryspy_obj = CrystRef.from_cif(str(obj))
        else:
            obj = self.__widg_central._object
            cryspy_obj = CrystRef.from_cif(obj.to_cif())
            if cryspy_obj is None:
                if mode_orig != "cryspy":
                    flag = False
        if cryspy_obj is None:
            cryspy_obj = CrystRef()
        if flag:
            self.__mode = "magref"
            cryspy_obj.file_input = self.__f_file
            self.__widg_central.form_cpanel(cryspy_obj)

    def save(self):
        f_name = self.__f_file
        if f_name is None:
            f_dir_data = self.__f_dir_data
            f_name = os.path.join(f_dir_data, "main.rcif")
        if self.__mode == "star":
            star_obj = self.__widg_central._star
            with open(f_name, "w") as fid:
                fid.write(str(star_obj))
        elif self.__mode in ["cryspy", "rhochi", "mem", "magref"]:
            cryspy_obj = self.__widg_central._object
            if cryspy_obj is None:
                return
            cryspy_obj.file_input = f_name
            if self.__mode == "cryspy":
                with open(f_name, "w") as fid:
                    fid.write(cryspy_obj.to_cif())
            else:
                cryspy_obj.save_to_file(f_name)
        return

    def save_as(self):
        f_name, okPressed = QtWidgets.QFileDialog.getSaveFileName(self, 'Select a file:', self.__f_dir_data,
                                                                  "Rcif files (*.rcif)")
        if not (okPressed):
            return None
        self.__f_file = f_name
        self.__f_dir_data = os.path.dirname(f_name)
        os.chdir(os.path.dirname(f_name))
        self.save()

    def open_folder(self):
        os.startfile(self.__f_dir_data)

    def open(self):
        f_name, okPressed = QtWidgets.QFileDialog.getOpenFileName(self, 'Select a cif file:', self.__f_dir_data,
                                                                  "All files (*.*)")
        if not (okPressed):
            return None
        self.__widg_central._star = None
        self.__widg_central._object = None
        self.__f_file = f_name
        self.__f_dir_data = os.path.dirname(f_name)
        os.chdir(os.path.dirname(f_name))
        mode_orig = str(self.__mode)
        self.to_star()
        if mode_orig == "cryspy":
            self.to_cryspy()
        elif mode_orig == "rhochi":
            self.to_rhochi()
        elif mode_orig == "mem":
            self.to_mem()
        elif mode_orig == "magref":
            self.to_magref()


class cwidget(QtWidgets.QSplitter):
    def __init__(self, width, height):
        self.info_width = width
        self.info_height = height
        super(cwidget, self).__init__()
        self._star = None
        self._object = None
        self.additional_thread = MyThread()
        self.additional_thread.signal_end.connect(self.end_of_calc_in_thread)
        self.additional_thread.signal_refresh.connect(self.refresh_wind)
        self.additional_thread.signal_start.connect(
            self.start_of_calc_in_thread)
        
        self.init_widget()
        self.additional_thread.w_cpanel = self.w_cpanel

    def init_widget(self):
        """
        make central layout
        """
        width_m_1 = self.info_width / 6.
        width_m_2 = (5. * self.info_width) / 6.
        width_v_1 = self.info_height * 0.75
        width_v_2 = self.info_height * 0.25

        self.w_cpanel = QtWidgets.QWidget()
        self.w_cpanel.setLayout(QtWidgets.QVBoxLayout())

        self.w_cpanel.setAutoFillBackground(True)
        self.w_cpanel.setStyleSheet("background-color:light gray;")

        self.w_output = QtWidgets.QLabel()
        self.w_output.setSizePolicy(QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding,
                                                          QtWidgets.QSizePolicy.Expanding))
        self.w_output.setFont(QtGui.QFont("Courier", 8, QtGui.QFont.Normal))
        self.w_output.setTextInteractionFlags(QtCore.Qt.TextSelectableByMouse)
        self.w_output.setAlignment(QtCore.Qt.AlignTop)
        self.w_output.setWordWrap(True)

        self.w_tab = QtWidgets.QTabWidget()

        self.w_11 = QtWidgets.QWidget()
        self.w_11.setLayout(QtWidgets.QVBoxLayout())
        self.w_12 = QtWidgets.QWidget()
        self.w_12.setLayout(QtWidgets.QVBoxLayout())
        self.w_13 = QtWidgets.QWidget()
        self.w_13.setLayout(QtWidgets.QVBoxLayout())
        self.w_2 = QtWidgets.QWidget()
        self.w_2.setLayout(QtWidgets.QVBoxLayout())
        self.w_3 = QtWidgets.QWidget()
        self.w_3.setLayout(QtWidgets.QVBoxLayout())

        layout_1 = QtWidgets.QHBoxLayout()
        layout_1.addWidget(self.w_11)
        layout_1.addWidget(self.w_12)
        layout_1.addWidget(self.w_13)
        w_1 = QtWidgets.QWidget()
        w_1.setLayout(layout_1)

        self.w_tab.addTab(w_1, "Parameters")
        self.w_tab.addTab(self.w_2, "RCIF")
        self.w_tab.addTab(self.w_3, "Optional")

        area = QtWidgets.QScrollArea()
        area.setWidgetResizable(True)
        area.setWidget(self.w_output)

        w_0 = QtWidgets.QWidget()
        w_0.setLayout(QtWidgets.QVBoxLayout())
        w_0.layout().addWidget(self.w_tab)

        lay_setting = QtWidgets.QHBoxLayout()
        self.w_loader = QtWidgets.QRadioButton(" calculations")

        self.w_loader.setChecked(False)

        lay_setting.addWidget(self.w_loader)
        lay_setting.addStretch(1)
        cb_1 = QtWidgets.QCheckBox()
        cb_1.setCheckState(2)
        lay_setting.addWidget(cb_1)
        cb_2 = QtWidgets.QCheckBox()
        cb_2.setCheckState(2)
        lay_setting.addWidget(cb_2)
        cb_3 = QtWidgets.QCheckBox()
        cb_3.setCheckState(2)
        lay_setting.addWidget(cb_3)
        w_0.layout().addLayout(lay_setting)

        cb_1.stateChanged.connect(lambda _1: show_widget(self.w_11, _1))
        cb_2.stateChanged.connect(lambda _1: show_widget(self.w_12, _1))
        cb_3.stateChanged.connect(lambda _1: show_widget(self.w_13, _1))

        w_out = QtWidgets.QWidget()
        w_out.setLayout(QtWidgets.QHBoxLayout())
        w_out.layout().addWidget(area)

        splitter_vert = QtWidgets.QSplitter()
        splitter_vert.setOrientation(QtCore.Qt.Vertical)
        splitter_vert.addWidget(w_0)
        splitter_vert.addWidget(w_out)
        splitter_vert.setSizes([width_v_1, width_v_2])

        self.addWidget(self.w_cpanel)
        self.addWidget(splitter_vert)
        self.setSizes([width_m_1, width_m_2])
        self.form_cpanel()

    def form_cpanel(self, obj=None):
        lay_cpanel = self.w_cpanel.layout()
        del_layout(lay_cpanel)

        if obj is None:
            obj = self._object
        if obj is not None:
            self._object = obj 
            w1 = QtWidgets.QTreeWidget()
            w1.setSizePolicy(QtWidgets.QSizePolicy(
                QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding))
            w1.setColumnCount(1)
            w1.setHeaderHidden(True)

            wi = QtWidgets.QTreeWidgetItem()
            s_name = type(obj).__name__
            wi.setText(0, f"{s_name:}")
            wi._object = obj
            w1.addTopLevelItem(wi)
            if isinstance(obj, Global):
                self._star = obj
                for _loop in obj.loops:
                    wii = QtWidgets.QTreeWidgetItem()
                    if _loop.name != "":
                        wii.setText(0, f"loop_{_loop.name:}{_loop.prefix:}")
                    else:
                        wii.setText(0, f"loop{_loop.prefix:}")
                    wii._object = _loop
                    wii.setToolTip(0, _loop.__doc__)
                    wi.addChild(wii)
                for _item in obj.items:
                    wii = QtWidgets.QTreeWidgetItem()
                    wii.setText(0, _item.name)
                    wii._object = _item
                    wii.setToolTip(0, _item.__doc__)
                    wi.addChild(wii)
                for _data in obj.datas:
                    wi = QtWidgets.QTreeWidgetItem()
                    wi.setText(0, f"data_{_data.name}")
                    wi.setToolTip(0, _data.__doc__)
                    wi._object = _data
                    for _loop in _data.loops:
                        wii = QtWidgets.QTreeWidgetItem()
                        if _loop.name != "":
                            wii.setText(
                                0, f"loop_{_loop.name:}{_loop.prefix:}")
                        else:
                            wii.setText(0, f"loop{_loop.prefix:}")
                        wii._object = _loop
                        wii.setToolTip(0, _loop.__doc__)
                        wi.addChild(wii)
                    for _item in _data.items:
                        wii = QtWidgets.QTreeWidgetItem()
                        wii.setText(0, _item.name)
                        wii._object = _item
                        wii.setToolTip(0, _item.__doc__)
                        wi.addChild(wii)

                    w1.addTopLevelItem(wi)
            else:
                self._object = obj
                l_data_obj = obj.mandatory_objs + obj.optional_objs
                for _data_obj in l_data_obj:
                    wi = QtWidgets.QTreeWidgetItem()
                    wi._object = _data_obj
                    wi.setToolTip(0, _data_obj.__doc__)
                    s_name = type(_data_obj).__name__
                    if isinstance(_data_obj, DataConstr):
                        wi.setText(0, f"{s_name:}: {_data_obj.data_name:}")
                    elif isinstance(_data_obj, LoopConstr):
                        wi.setText(0, f"{s_name:}: {_data_obj.loop_name:}")
                    elif isinstance(_data_obj, ItemConstr):
                        wi.setText(0, f"{s_name:}")

                    if isinstance(_data_obj, DataConstr):
                        l_loop_item_obj = _data_obj.mandatory_objs + \
                            _data_obj.optional_objs+_data_obj.internal_objs

                        for _li_obj in l_loop_item_obj:
                            wii = QtWidgets.QTreeWidgetItem()
                            flag_2 = False
                            if isinstance(_li_obj, ItemConstr):
                                s_li_obj = type(_li_obj).__name__
                                flag_2 = True
                            elif isinstance(_li_obj, LoopConstr):
                                s_li_obj = f"{type(_li_obj).__name__:}: {_li_obj.loop_name:}"
                                flag_2 = True
                            if flag_2:
                                wii.setText(0, s_li_obj)
                                wii._object = _li_obj
                                wii.setToolTip(0, _li_obj.__doc__)
                                wi.addChild(wii)

                                l_int_attr = _li_obj.INTERNAL_ATTRIBUTE
                                if l_int_attr is None:
                                    l_int_attr = []
                                for _int_attr in l_int_attr:
                                    _int_obj = getattr(_li_obj, _int_attr)
                                    if ((isinstance(_int_obj, ItemConstr)) | (isinstance(_int_obj, LoopConstr))):
                                        wiii = QtWidgets.QTreeWidgetItem()
                                        wiii.setText(0, f"{_int_attr:}")
                                        wiii._object = _int_obj
                                        wiii.setToolTip(0, _int_obj.__doc__)
                                        wii.addChild(wiii)
                    else:
                        l_int_attr = _data_obj.INTERNAL_ATTRIBUTE
                        if l_int_attr is None:
                            l_int_attr = []
                        for _int_attr in l_int_attr:
                            _int_obj = getattr(_data_obj, _int_attr)
                            if (isinstance(_int_obj, ItemConstr)) | (isinstance(_int_obj, LoopConstr)):
                                wii = QtWidgets.QTreeWidgetItem()
                                wii.setText(0, f"{_int_attr:}")
                                wii._object = _int_obj
                                wii.setToolTip(0, _int_obj.__doc__)
                                wi.addChild(wii)

                    w1.addTopLevelItem(wi)

            w1.itemClicked.connect(self.onItemClicked)
            w1.expandAll()
            lay_cpanel.addWidget(w1)

    @QtCore.pyqtSlot(QtWidgets.QTreeWidgetItem, int)
    def onItemClicked(self, it, col):
        self.form_tab(it._object)

    def form_tab(self, obj):
        layout_11 = self.w_11.layout()
        layout_12 = self.w_12.layout()
        layout_13 = self.w_13.layout()
        layout_2 = self.w_2.layout()
        layout_3 = self.w_3.layout()
        w_output = self.w_output
        del_layout(layout_11)
        del_layout(layout_12)
        del_layout(layout_13)
        del_layout(layout_2)
        del_layout(layout_3)

        thread = self.additional_thread

        flag_done = False
        if (FLAG_MAGREF & (not(flag_done))):
            if isinstance(obj, MeasInelasticL):
                w_for_meas_inelasticl(obj, layout_11, layout_12,
                                  layout_13, layout_2, layout_3, w_output, thread)
                flag_done = True

        if (FLAG_MEM & (not(flag_done))):
            pass

        if (FLAG_CRYSPY & (not(flag_done))):
            if isinstance(obj, PdPeakL):
                w_for_pd_peak(obj, layout_11, layout_12, layout_13,
                              layout_2, layout_3, w_output, thread)
                flag_done = True
            elif isinstance(obj, PdMeasL):
                w_for_pd_meas(obj, layout_11, layout_12, layout_13,
                              layout_2, layout_3, w_output, thread)
                flag_done = True
            elif isinstance(obj, PdProcL):
                w_for_pd_proc(obj, layout_11, layout_12, layout_13,
                              layout_2, layout_3, w_output, thread)
                flag_done = True
            elif isinstance(obj, DiffrnReflnL):
                w_for_diffrn_refln(obj, layout_11, layout_12, layout_13,
                                   layout_2, layout_3, w_output, thread)
                flag_done = True
            elif isinstance(obj, Pd2dMeas):
                w_for_pd2d_meas(obj, layout_11, layout_12, layout_13,
                                layout_2, layout_3, w_output, thread)
                flag_done = True
            elif isinstance(obj, Pd2dProc):
                w_for_pd2d_proc(obj, layout_11, layout_12, layout_13,
                                layout_2, layout_3, w_output, thread)
                flag_done = True
            elif isinstance(obj, Pd):
                w_for_pd(obj, layout_11, layout_12, layout_13,
                         layout_2, layout_3, w_output, thread)
                flag_done = True

        if not(flag_done):
            if isinstance(obj, GlobalConstr):
                w_for_global_constr(obj, layout_11, layout_12,
                                    layout_13, layout_2, layout_3, w_output, thread)
            elif isinstance(obj, DataConstr):
                w_for_data_constr(obj, layout_11, layout_12,
                                  layout_13, layout_2, layout_3, w_output, thread)
            elif isinstance(obj, LoopConstr):
                w_for_loop_constr(obj, layout_11, layout_12,
                                  layout_13, layout_2, layout_3, w_output, thread)
            elif isinstance(obj, ItemConstr):
                w_for_item_constr(obj, layout_11, layout_12,
                                  layout_13, layout_2, layout_3, w_output, thread)
            elif isinstance(obj, Global):
                pass
            elif isinstance(obj, Data):
                pass
            elif isinstance(obj, Loop):
                w_for_loop(obj, layout_11, layout_12,
                           layout_13, layout_2, layout_3, w_output, thread)
            elif isinstance(obj, Item):
                w_for_item(obj, layout_11, layout_12,
                           layout_13, layout_2, layout_3, w_output, thread)

    def start_of_calc_in_thread(self):
        self.w_loader.setChecked(True)
        self.w_cpanel.setStyleSheet("background-color:white;")

        ls_out = [f"The calculations are running.\n"]
        self.w_output.setText("\n".join(ls_out))

    def end_of_calc_in_thread(self):
        self.w_loader.setChecked(False)
        self.w_cpanel.setStyleSheet("background-color:light gray;")

        thread = self.additional_thread
        function = thread.function
        function_name = function.__name__
        output = thread.output
        ls_out = []
        ls_out.append(
            f"The function '{function_name:}' is perfomed.\n\nResult is \n")
        ls_out.append(str(output))
        self.w_output.setText("\n".join(ls_out))
        self.form_cpanel()

    def refresh_wind(self):
        self.form_cpanel()


def main_w(l_arg=[]):
    app = QtWidgets.QApplication(l_arg)
    app.setStyle(QtWidgets.QStyleFactory.create('Fusion'))
    if len(l_arg) >= 2:
        f_dir_data = os.path.abspath(l_arg[1])
    else:
        f_dir_data = os.getcwd()
    main_wind_1 = cbuilder(f_dir_data)
    sys.exit(app.exec_())


if __name__ == '__main__':
    l_arg = sys.argv
    main_w(l_arg)



import os
import sys
import pycifstar

from cryspy_editor.main_window import main_w



l_argv = sys.argv
flag = True

print("Programm 'cryspy_editor'.\n")
if flag:
    main_w(l_argv)
else:
    pass
print("That's all")


import os
import numpy
from PyQt5 import QtWidgets, QtGui, QtCore
from cryspy_editor.b_rcif_to_cryspy import L_ITEM_CLASS, L_LOOP_CLASS, L_DATA_CLASS
from .FUNCTIONS import show_info, get_layout_method_help,  make_qtablewidget_for_data_constr, show_widget, add_mandatory_optional_obj

from cryspy.common.cl_item_constr import ItemConstr
from cryspy.common.cl_loop_constr import LoopConstr


def w_for_data_constr(obj, layout_11, layout_12, layout_13, layout_2, layout_3, w_output, thread):


    layout_11.addWidget(QtWidgets.QLabel("Data name:"))
    _l_e_dist = QtWidgets.QLineEdit()
    _l_e_dist.setText(obj.data_name)
    _l_e_dist.editingFinished.connect(lambda : setattr(obj, "data_name", _l_e_dist.text()))
    layout_11.addWidget(_l_e_dist)

    layout_11.addWidget(QtWidgets.QLabel("Defined attributes:"))
    _widg_table = make_qtablewidget_for_data_constr(obj, w_output, thread)
    layout_11.addWidget(_widg_table)

    _b_info = QtWidgets.QPushButton("info")
    _b_info.clicked.connect(lambda : show_info(obj, w_output))
    layout_11.addWidget(_b_info)

    lay_left_2 = add_mandatory_optional_obj(obj, w_output, thread)
    layout_12.addLayout(lay_left_2)

    lay_left_3 = get_layout_method_help(obj, w_output, thread)
    layout_13.addLayout(lay_left_3)


    _text_edit = QtWidgets.QTextEdit()
    _text_edit.setText(obj.to_cif())
    layout_2.addWidget(_text_edit)

    return


def create_obj(widg, obj):
    ls_out = ["Enter the item:"]
    l_class = obj.mandatory_classes+obj.optional_classes
    
    n_mandatory = len(obj.mandatory_classes)

    l_h = []
    for _class in l_class:
        if isinstance(_class, ItemConstr):
            l_h.append(_class.PREFIX, _class)
        elif isinstance(_class, LoopConstr):
            l_h.append(_class.ITEM_CLASS.PREFIX, _class)

    ls_out.extend([f"{_i+1}: {_h[0]:}" for _i, _h in enumerate(l_h)])

    text, ok = QtWidgets.QInputDialog.getText(widg, 'Input Dialog',
                            "\n".join(ls_out))
    if not(ok):
        return None

    _ind = int(text)-1
    item_class = l_h[_ind][1]
    item = item_class()
    if _ind >= n_mandatory:
        obj.optional_objs.append(item)
    else:
        obj.mandatory_objs.append(item)

def create_items(widg, obj):
    ls_out = ["Enter the item:"]
    ls_out.extend([f"{_i+1}: {_item.PREFIX:}" for _i, _item in enumerate(L_ITEM_CLASS)])

    text, ok = QtWidgets.QInputDialog.getText(widg, 'Input Dialog',
                            "\n".join(ls_out))
    if not(ok):
        return None

    _ind = int(text)-1
    item_class = L_ITEM_CLASS[_ind]
    item = item_class()
    if not(item_class in obj.mandatory_classes+obj.optional_classes):
        obj.optional_classes.append(item_class)
    obj.optional_objs.append(item)
    

def create_loop(widg, obj):
    ls_out = ["Enter the loop:"]
    ls_out.extend([f"{_i+1}: {_item.ITEM_CLASS.PREFIX:}" for _i, _item in enumerate(L_LOOP_CLASS)])

    text, ok = QtWidgets.QInputDialog.getText(widg, 'Input Dialog',
                            "\n".join(ls_out))
    if not(ok):
        return None

    _ind = int(text)-1
    loop_class = L_LOOP_CLASS[_ind]
    loop = loop_class()
    if not(loop_class in obj.mandatory_classes+obj.optional_classes):
        obj.optional_classes.append(loop_class)
    obj.optional_objs.append(loop)



import os
import sys

import pycifstar

from PyQt5 import QtWidgets
from PyQt5 import QtGui
from PyQt5 import QtCore
 
import matplotlib
import matplotlib.backends.backend_qt5agg
import matplotlib.figure 
import matplotlib.pyplot

import numpy
import math
import scipy
import scipy.optimize


class ccomunicate(QtCore.QObject):
    sig_m = QtCore.pyqtSignal()
    sig_l_tth = QtCore.pyqtSignal()
    sig_l_phi = QtCore.pyqtSignal()
    sig_l_redraw = QtCore.pyqtSignal()





class cmodel(QtCore.QObject):
    """
    class to control changings in the widget and to interact with data
    """
    def __init__(self, x, y, z):
        super(cmodel, self).__init__()
        self.data_matrix = z
        self.data_range_x = x
        self.data_range_y = y
        
        self.min_x = numpy.nan
        self.max_x = numpy.nan
        self.min_y = numpy.nan
        self.max_y = numpy.nan
        self.min_z_u = numpy.nan
        self.max_z_u = numpy.nan
        self.reset_limit()
        
    def reset_limit(self):
        self.min_x = numpy.nanmin(self.data_range_x)
        self.max_x = numpy.nanmax(self.data_range_x)
        self.min_y = numpy.nanmin(self.data_range_y)
        self.max_y = numpy.nanmax(self.data_range_y)        
        
        
        self.min_z_u = numpy.nanmin(self.data_matrix)
        self.max_z_u = numpy.nanmax(self.data_matrix)

        
    def index_to_value(self, ind_x, ind_y):
        iint_x = int(round(ind_x))
        iint_y = int(round(ind_y))
        val_x = self.data_range_x[iint_x]
        val_y = self.data_range_y[iint_y]
        return val_x, val_y
        
    def value_to_index(self, val_x, val_y):
        min_x = numpy.nanmin(self.data_range_x)
        max_x = numpy.nanmax(self.data_range_x)
        min_y = numpy.nanmin(self.data_range_y)
        max_y = numpy.nanmax(self.data_range_y)
        ind_x = int(round((self.data_range_x.size-1)*(val_x-min_x)*1./(max_x-min_x)))
        ind_y = int(round((self.data_range_y.size-1)*(val_y-min_y)*1./(max_y-min_y)))
        return ind_x, ind_y



class cwidg_central(QtWidgets.QWidget):
    def __init__(self):
        super(cwidg_central, self).__init__()
        self.init_layout_central()
        self.setLayout(self.layout_central)
        
 
    def init_layout_central(self):
        
        lay_main = QtWidgets.QHBoxLayout()
        
        self.graph_m = cgraph_matrix(self)
        self.graph_m.figure.canvas.mpl_connect("button_press_event", self.onclick_matrix)

        def _temp_func(_graph_m, label, val):
            if label == "min":
                _graph_m.image.set_clim(vmin=val)
            elif label == "max":
                _graph_m.image.set_clim(vmax=val)
            _graph_m.draw()
            return

        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        lay_right = QtWidgets.QVBoxLayout()
        _l_1 = QtWidgets.QLabel("z_min:")
        self._l_e_z_min = QtWidgets.QLineEdit()
        self._l_e_z_min.editingFinished.connect(lambda : _temp_func(self.graph_m, "min", float(str(self._l_e_z_min.text()))))

        _l_2 = QtWidgets.QLabel("z_max:")
        self._l_e_z_max = QtWidgets.QLineEdit()
        self._l_e_z_max.editingFinished.connect(lambda : _temp_func(self.graph_m, "max", float(str(self._l_e_z_max.text()))))
        
        self._l_coord = QtWidgets.QLabel()

        _b_gm = QtWidgets.QPushButton("matrix to clipboard")
        _b_gm.clicked.connect(self.give_matrix)

        lay_right.addWidget(_l_1)
        lay_right.addWidget(self._l_e_z_min)
        lay_right.addWidget(_l_2)
        lay_right.addWidget(self._l_e_z_max)
        lay_right.addWidget(self._l_coord)
        lay_right.addStretch(1)
        lay_right.addWidget(_b_gm)

        _l_1.setSizePolicy(sizePolicy)
        self._l_e_z_min.setSizePolicy(sizePolicy)
        _l_2.setSizePolicy(sizePolicy)
        self._l_e_z_max.setSizePolicy(sizePolicy)
        _b_gm.setSizePolicy(sizePolicy)


        """
        splitter_h = QtWidgets.QSplitter(QtCore.Qt.Horizontal)
        splitter_v = QtWidgets.QSplitter(QtCore.Qt.Vertical)
        splitter_v.addWidget(self.graph_l_tth)
        splitter_v.addWidget(self.graph_l_phi)
        splitter_h.addWidget(self.graph_m)
        splitter_h.addWidget(splitter_v)
        lay_cp = self.control_panel()
        lay_1 = QtWidgets.QVBoxLayout()
        lay_1.addWidget(splitter_h)
        lay_1.addLayout(lay_cp)
        """
        lay_main.addWidget(self.graph_m)
        lay_main.addLayout(lay_right)
        self.layout_central = lay_main

    def give_matrix(self):
        val_x = self.graph_m.val_x
        val_y = self.graph_m.val_y
        val_z = self.graph_m.val_z

        ls_out = []
        ls_out.append(f"{val_y.size:15} " + " ".join([f"{_:15.2f}" for _ in val_x]))
        for _y, _l_z in zip(val_y, val_z):
            sline = f"{_y:15.2f} "+" ".join(["           None" if numpy.isnan(_) else f"{_:15.1f}" for _ in _l_z])
            ls_out.append(sline)
        
        cb = QtWidgets.QApplication.clipboard()
        cb.clear(mode=cb.Clipboard)
        cb.setText("\n".join(ls_out), mode=cb.Clipboard)

    def onclick_matrix(self, event):
        ind_x = int(round(event.xdata))
        ind_y = int(round(event.ydata))
        val_x, val_y = self.model.index_to_value(ind_x, ind_y)
        str_x = "{:.3f}".format(val_x)
        str_y = "{:.3f}".format(val_y)
    
        val_z = None
        val_z = self.model.data_matrix[ind_y, ind_x]

        str_z = "{:.3f}".format(val_z)
        
        if event.button == 1:
            self._l_coord.setText(f"x: {val_x:.2f}\ny: {val_y:.2f}\nz: {val_z:.2f}")
        elif event.button == 2:
            self.graph_m.set_lim(0, len(self.graph_m.val_x)-1, "x")
            self.graph_m.set_lim(0, len(self.graph_m.val_y)-1, "y")
            self.graph_m.lim_flag = True
            
        elif event.button == 3:
            if self.graph_m.lim_flag:
                self.graph_m.lim_flag = False
                self.graph_m.xlim_1 = ind_x
                self.graph_m.ylim_1 = ind_y
            else:
                self.graph_m.lim_flag = True
                self.graph_m.set_lim(None, ind_x, "x")
                self.graph_m.set_lim(None, ind_y, "y")
    
    def plot_matrix(self, x, y, z):
        self.model = cmodel(x, y, z)
        self.graph_m.set_x(x)
        self.graph_m.set_y(y)
        self.graph_m.set_z(z)
        
        try:
            flag = numpy.logical_not(numpy.isnan(z))
            self._l_e_z_min.setText(f"{min(z[flag]):.0f}")
            self._l_e_z_max.setText(f"{max(z[flag]):.0f}")
        except:
            pass
        
        self.graph_m.set_data_to_graph()

    def change_value(self, value):
        print(value)
        return


       
class cgraph(matplotlib.backends.backend_qt5agg.FigureCanvasQTAgg):
 
    def __init__(self, parent=None, width=5, height=4, dpi=100):
        self.fig = matplotlib.figure.Figure(figsize=(width, height), dpi = dpi)
        self.fig.subplots_adjust(left = 0.12,
                            right = 0.90,
                            top = 0.90,
                            bottom = 0.12,
                            wspace = 0.0,
                            hspace = 0.0)
        super(cgraph, self).__init__(self.fig)
        self.xlim_1 = None
        self.xlim_2 = None
        self.ylim_1 = None
        self.ylim_2 = None
        self.zlim_1 = None
        self.zlim_2 = None
        self.lim_flag = True
        
        self.val_x = None
        self.val_y = None
        self.val_z = None
        
        self.info_press = (False, False)
        self.control = parent
        self.figure = self.fig
        self.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        self.ax_pri = self.fig.add_subplot(111)
        #fig.canvas.mpl_connect("button_press_event", self.onclick)
        
        
    def onclick(self, event):
        print("x position: ", event.xdata)
        print("y position: ", event.ydata)
        
    def set_x(self, np_x):
        self.val_x = np_x
        
    def set_y(self, np_y):
        self.val_y = np_y
        
    def set_z(self, np_z):
        self.val_z = np_z
    
    def set_lim(self, xlim_1, xlim_2, ax):
        
        if xlim_1 != None:
            if ax == "x":
                self.xlim_1 = xlim_1
            elif ax == "y":
                self.ylim_1 = xlim_1
            elif ax == "z":
                self.zlim_1 = xlim_1
        if xlim_2 != None:
            if ax == "x":
                self.xlim_2 = xlim_2
            elif ax == "y":
                self.ylim_2 = xlim_2
            elif ax == "z":
                self.zlim_2 = xlim_2
        if ax == "x":
            if (self.xlim_1 != None) & (self.xlim_2 != None):
                x_min = min([self.xlim_1, self.xlim_2])
                x_max = max([self.xlim_1, self.xlim_2])
                self.ax_pri.set_xlim([x_min, x_max])
                self.draw()        
        elif ax == "y":
            if (self.ylim_1 != None) & (self.ylim_2 != None):
                x_min = min([self.ylim_1, self.ylim_2])
                x_max = max([self.ylim_1, self.ylim_2])
                self.ax_pri.set_ylim([x_min, x_max])
                self.draw()        
        elif ax == "z":
            if (self.zlim_1 != None) & (self.zlim_2 != None):
                x_min = min([self.zlim_1, self.zlim_2])
                x_max = max([self.zlim_1, self.zlim_2])
                self.image.set_clim(vmin = x_min, vmax = x_max)
                self.draw()        

     

class cgraph_matrix(cgraph):
    def __init__(self, parent=None):
        super(cgraph_matrix, self).__init__(cgraph)
        self.zlim_1 = None
        self.zlim_2 = None
        self.fig.subplots_adjust(left = 0.02,
                            right = 0.95,
                            top = 0.95,
                            bottom = 0.02,
                            wspace = 0.0,
                            hspace = 0.0)
        
        
        
    def onclick(self, event):
        ind_tth = int(event.xdata)
        ind_phi = int(event.ydata)
        if event.button == 1:
            self.val_x_inline = self.val_z[ind_phi, :]
            self.val_y_inline = self.val_z[:, ind_tth]
            print("ttheta: {:.2f}, phi: {:.2f} ".format(self.val_x[ind_tth], 
                           self.val_y[ind_phi]))
            self.sig.sig_m.emit()
        elif event.button == 2:
            self.set_lim(0, len(self.val_x)-1, "x")
            self.set_lim(0, len(self.val_y)-1, "y")
            self.lim_flag = True
            self.sig.sig_l_redraw.emit()
            
        elif event.button == 3:
            ind_tth = int(event.xdata)
            ind_phi = int(event.ydata)
            if self.lim_flag:
                print("Step 1")
                self.lim_flag = False
                self.xlim_1 = ind_tth
                self.ylim_1 = ind_phi
            else:
                print("Step 2",[self.xlim_1, ind_tth], [self.ylim_1, ind_phi])
                self.lim_flag = True
                self.set_lim(None, ind_tth, "x")
                self.set_lim(None, ind_phi, "y")
                self.sig.sig_l_redraw.emit()

        
    def set_data_to_graph(self):
        self.ax_pri.cla()
        if (self.zlim_1 != None) and (self.zlim_2 != None):
            image = self.ax_pri.imshow(self.val_z, aspect = 'auto', origin = 'lower', vmin = self.zlim_1, vmax = self.zlim_2)
        else:
            image = self.ax_pri.imshow(self.val_z, aspect = 'auto', origin = 'lower')
        #self.ax_pri.set_axis_off()
        #numpy.arange(len(self.val_x)),self.val_x
        self.ax_pri.set_xticks([])
        
        self.ax_pri.set_yticks([])
        if ((self.xlim_1 != None)&(self.xlim_2 != None)):
            x_min = min([self.xlim_1 , self.xlim_2])
            x_max = max([self.xlim_1 , self.xlim_2])
            self.ax_pri.set_xlim([x_min, x_max])
        if ((self.ylim_1 != None)&(self.ylim_2 != None)):
            y_min = min([self.ylim_1 , self.ylim_2])
            y_max = max([self.ylim_1 , self.ylim_2])
            self.ax_pri.set_ylim([y_min, y_max])
        #self.ax_pri.plot(self.data_x, self.data_y, "-")
        #self.ax_pri.matshow(self.np_int)
        #self.ax_pri.set_xticks(self.np_tth)
        #self.ax_pri.set_yticks(self.np_phi)
        #print self.np_phi.size
        #print self.np_tth.size
        #print self.np_int.shape
        #self.ax_pri.plot(self.data_x, self.data_x, "k-", linewidth=1.0)    
        #self.ax_pri.errorbar(self.data_x, self.data_y, yerr = self.data_sy, ecolor = col_1, fmt='o', color=col_1, linewidth = 0.5)
        self.image = image
        self.draw()

        



import os
import sys
 
from PyQt5 import QtWidgets
from PyQt5 import QtGui
from PyQt5 import QtCore
import numpy
import matplotlib
import matplotlib.backends.backend_qt5agg
import matplotlib.figure 
import matplotlib.pyplot
 

def read_data(file_name):

    fid = open(file_name,'r')
    l_cont = fid.readlines()
    fid.close()

    lmat, mat = [], []
    for hh in l_cont:
        if hh.strip() == "":
            if mat != []:
                lmat.append(mat)
                mat = []
        else:
            mat.append(hh)
    if mat != []:
        lmat.append(mat)
        mat = []
    
    ll_int, l_ang1, l_ang2 = [], [], []
    for mat in lmat:
        ang1 = [float(hh) for hh in mat[0].split()[1:]]
        l_int, ang2 = [], []
        for hh in mat[1:]:
            lhelp = hh.split()
            ang2.append(float(lhelp[0]))
            l_int.append([float(hh2) if hh2 != "None" else None for hh2 in lhelp[1:]])
        ll_int.append(l_int)
        l_ang1.append(ang1)
        l_ang2.append(ang2)

    tth = numpy.array(l_ang1[0], dtype=float)
    phi = numpy.array(l_ang2[0], dtype=float)
    int_s_e = numpy.array(ll_int[0], dtype=float).transpose()
    int_s_m = numpy.array(ll_int[1], dtype=float).transpose()
    sint_s = numpy.array(ll_int[2], dtype=float).transpose()
    int_d_e = numpy.array(ll_int[3], dtype=float).transpose()
    int_d_m = numpy.array(ll_int[4], dtype=float).transpose()
    sint_d = numpy.array(ll_int[5], dtype=float).transpose()
    return tth, phi, int_s_e, int_s_m, sint_s, int_d_e, int_d_m, sint_d


class cwind_central(QtWidgets.QMainWindow):
    def __init__(self):
        super(cwind_central, self).__init__()
        self.title = "program 'Graph'"
        self.setWindowTitle(self.title)
        widg_central = cwidg_central()
        self.setCentralWidget(widg_central)
        self.show()



class cwidg_central(QtWidgets.QWidget):
    def __init__(self):
        super(cwidg_central, self).__init__()
        self.init_layout_central()
        self.setLayout(self.layout_central)
 
    def init_layout_central(self):
        
        lay_main = QtWidgets.QVBoxLayout()
        self.graph = Graph(self, width=5, height=4)
        

        lay_1 = QtWidgets.QVBoxLayout()
        lay_1.addWidget(self.graph)

        lay_main.addLayout(lay_1)

        
        self.layout_central = lay_main
        
    def plot_file(self, ffig_full):
        self.graph.ax_pri.cla()
        tth, phi, int_s_e, int_s_m, sint_s, int_d_e, int_d_m, sint_d = read_data(ffig_full)

        i_s_e_1d = numpy.where(numpy.isnan(int_s_e), 0., int_s_e).sum(axis=1)
        i_s_m_1d = numpy.where(numpy.isnan(int_s_e), 0., int_s_m).sum(axis=1)
        si_s_e_1d = ((numpy.where(numpy.isnan(int_s_e), 0., sint_s)**2).sum(axis=1))**0.5
        i_d_e_1d = numpy.where(numpy.isnan(int_d_e), 0., int_d_e).sum(axis=1)
        i_d_m_1d = numpy.where(numpy.isnan(int_d_e), 0., int_d_m).sum(axis=1)
        si_d_e_1d = ((numpy.where(numpy.isnan(int_d_e), 0., sint_d)**2).sum(axis=1))**0.5

        self.graph.data_1_x = []
        self.graph.data_1_y_exp = []
        self.graph.data_1_sy_exp = []
        self.graph.data_1_y_mod = []
        self.graph.data_1_y_diff = []
        
        
        self.graph.data_2_y_exp = []
        self.graph.data_2_sy_exp = []
        self.graph.data_2_y_mod = []
        self.graph.data_2_y_diff = []


        self.graph.data_3_x = tth
        self.graph.data_3_y_exp = i_s_e_1d
        self.graph.data_3_sy_exp = si_s_e_1d
        self.graph.data_3_y_mod = i_s_m_1d
        self.graph.data_3_y_diff = i_s_e_1d-i_s_m_1d
        

        self.graph.data_4_x = tth
        self.graph.data_4_y_exp = i_d_e_1d
        self.graph.data_4_sy_exp = si_d_e_1d
        self.graph.data_4_y_mod = i_d_m_1d
        self.graph.data_4_y_diff = i_d_e_1d-i_d_m_1d

        
        self.graph.set_data_to_graph()


class Graph(matplotlib.backends.backend_qt5agg.FigureCanvasQTAgg):
    
    def __init__(self, parent=None, width=5, height=4, dpi=100):
        fig = matplotlib.figure.Figure(figsize=(width, height), dpi = dpi)
        fig.subplots_adjust(left = 0.07,
                            right = 0.97,
                            top = 0.97,
                            bottom = 0.07,
                            wspace = 0.0,
                            hspace = 0.0)
        
        super(Graph, self).__init__(fig)
        
        self.info_press = (False, False)

        self.data_1_x = []
        self.data_1_y_exp = []
        self.data_1_sy_exp = []
        self.data_1_y_mod = []
        self.data_1_y_diff = []

        self.data_2_y_exp = []
        self.data_2_sy_exp = []
        self.data_2_y_mod = []
        self.data_2_y_diff = []


        self.data_3_x = []
        self.data_3_y_exp = []
        self.data_3_sy_exp = []
        self.data_3_y_mod = []
        self.data_3_y_diff = []
        
        self.data_4_x = []
        self.data_4_y_exp = []
        self.data_4_sy_exp = []
        self.data_4_y_mod = []
        self.data_4_y_diff = []

        self.data_5_x =  []
        
        self.control = parent
        self.figure = fig
        
        self.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        self.ax_pri = fig.add_subplot(111)
        
        fig.canvas.mpl_connect("button_press_event", self.onclick)
        
    def set_data_to_graph(self):
        col_1 = "#FF0000"
        col_2 = "#0000FF"
        col_3 = "#006400"
        lx = self.data_3_x
        ly1 = self.data_3_y_exp
        lsy1 = self.data_3_sy_exp
        y_up_min = min(ly1)
        ly1_mod = self.data_3_y_mod
        ly1_exp_mod = [-hh1+hh2+y_up_min for hh1, hh2 in zip(ly1, ly1_mod)]
        
        self.ax_pri.plot(lx, ly1_mod, "k-",linewidth=1.0)
        self.ax_pri.plot(lx, ly1_exp_mod, "k-",linewidth=0.5)
        self.ax_pri.errorbar(lx, ly1, yerr = lsy1, ecolor = col_1,  fmt='.', color=col_1, linewidth = 0.5)



        ly1_before = self.data_4_y_exp
        lsy1 = self.data_4_sy_exp
        y_down_max = max(ly1_before)
        y_down_min = min(ly1_before) - y_down_max + y_up_min
        ly1 = [hh - y_down_max + y_up_min  for hh in ly1_before]
        ly1_mod = [hh - y_down_max + y_up_min  for hh in self.data_4_y_mod]
        ly1_exp_mod = [-hh1+hh2 + y_down_min for hh1, hh2 in zip(ly1, ly1_mod)]
        
        #linestyle='-'
        self.ax_pri.plot(lx, ly1_mod, "k-",linewidth=1.0)
        self.ax_pri.plot(lx, ly1_exp_mod, "k-",linewidth=0.5)
        self.ax_pri.errorbar(lx, ly1, yerr = lsy1, ecolor = col_1, fmt='.', color=col_2, linewidth = 0.5)


        #ly1_before = self.data_4_y_exp
        #y_diff_max = max(ly1_before)
        #y_diff_min = min(ly1_before) - y_diff_max + y_down_min

        x_diff = max(lx)-min(lx)
        y_diff = max(self.data_4_y_exp) - y_down_min
        self.xlim_orig = (min(lx)-0.05*x_diff, max(lx)+0.05*x_diff)
        self.ylim_orig = (y_down_min-0.05*y_diff, max(self.data_3_y_exp)+0.05*y_diff)
        """
        self.ax_pri.plot(self.data_1_x, self.data_1_y_mod, "k-", linewidth=1.0)    
         
        self.ax_pri.errorbar(self.data_1_x, self.data_1_y_exp, yerr = self.data_1_sy_exp, ecolor = col_1, fmt='.', color=col_1, linewidth = 0.5)
        """ 
        self.draw()
        
    def onclick(self, event):
        if event.button == 3:
            
            if self.info_press == (False, False):
                self.info_press = (True, False)
                self.xlim = [event.xdata]
                self.ylim = [event.ydata]
            elif self.info_press == (True, False):
                self.info_press = (True, True)
                self.xlim.append(event.xdata)
                self.ylim.append(event.ydata)
            if self.info_press == (True, True):
                self.info_press = (False, False)
                xlim = (min(self.xlim), max(self.xlim))
                ylim = (min(self.ylim), max(self.ylim))
                self.ax_pri.set_xlim(xlim)
                self.ax_pri.set_ylim(ylim)
                self.xlim = []
                self.ylim = []
                self.draw()
        elif event.button == 2:
                self.info_press == (False, False)
                self.ax_pri.set_xlim(self.xlim_orig)
                self.ax_pri.set_ylim(self.ylim_orig)
                self.xlim = []
                self.ylim = []
                self.draw()
        else:
            self.info_press == (False, False)
            self.xlim = []
            self.ylim = []
            
 
if __name__ == '__main__':
    app = QtWidgets.QApplication(sys.argv)
    fname = "NaCaAlF_exp.out"
    ex = cwind_central()
    
    sys.exit(app.exec_())



import numpy
from PyQt5 import QtWidgets

from .FUNCTIONS import get_layout_rciftab_obj, del_layout
from .w_pd_proc import w_for_pd_proc
from .w_pd_meas import w_for_pd_meas
from .w_data_constr import w_for_data_constr
from .i_graph_mod_1d import cwidg_central as cwidg_pwd
from cryspy import PdMeasL, PdProcL, PdBackgroundL, PdPeakL

def w_for_pd(obj, layout_11, layout_12, layout_13, layout_2, layout_3, w_output, thread):
    w_for_data_constr(obj, layout_11, layout_12, layout_13, layout_2, layout_3, w_output, thread)
    del_layout(layout_3)

    f_proc, f_meas, f_bkgr, f_peak = False, False, False, False
    proc, meas, bkgr, peak = None, None, None, None
    meas = obj.meas
    f_meas = meas is not None
    bkgr = obj.background
    f_bkgr = bkgr is not None
    for int_obj in obj.internal_objs:
        if isinstance(int_obj, PdProcL):
            proc, f_proc = int_obj, True
        elif isinstance(int_obj, PdPeakL):
            peak, f_peak = int_obj, True
    if (f_proc & f_peak):
        np_x_1 = numpy.array(proc.ttheta, dtype = float)
        np_y_u_1 = numpy.array(proc.intensity_up, dtype = float)
        np_y_su_1 = numpy.array(proc.intensity_up_sigma, dtype = float)
        np_y_d_1 = numpy.array(proc.intensity_down, dtype = float)
        np_y_sd_1 = numpy.array(proc.intensity_down_sigma, dtype = float)
        np_y_b_1 = numpy.array(proc.intensity_bkg_calc, dtype = float)
        np_y_u_2 = numpy.array(proc.intensity_up_total, dtype = float)
        np_y_d_2 = numpy.array(proc.intensity_down_total, dtype = float)
        np_y_s_1 = numpy.array(proc.intensity, dtype = float)
        np_y_ss_1 = numpy.array(proc.intensity_sigma, dtype = float)
        np_y_m_1 = np_y_u_1 - np_y_d_1
        np_y_sm_1 = np_y_ss_1
        np_y_s_2 = numpy.array(proc.intensity_total, dtype = float)
        np_y_m_2 = numpy.array(proc.intensity_diff_total, dtype = float)

        np_x_2 = numpy.array(peak.ttheta, dtype = float)
        np_xysm_1 = numpy.vstack((np_x_1, np_y_s_1, np_y_ss_1, np_y_s_2)).transpose()
        np_xb_1 = numpy.vstack((np_x_1, 2*np_y_b_1)).transpose()
        w_s_1 = cwidg_pwd()
        w_s_1.plot_numpy_arrays(np_x_2, np_xysm_1, np_xb_1)

        np_xysm_1 = numpy.vstack((np_x_1, np_y_m_1, np_y_sm_1, np_y_m_2)).transpose()
        w_s_2 = cwidg_pwd()
        w_s_2.plot_numpy_arrays(np_x_2, np_xysm_1)

        np_xysm_1 = numpy.vstack((np_x_1, np_y_u_1, np_y_su_1, np_y_u_2)).transpose()
        np_xb_1 = numpy.vstack((np_x_1, np_y_b_1)).transpose()
        w_s_3 = cwidg_pwd()
        w_s_3.plot_numpy_arrays(np_x_2, np_xysm_1, np_xb_1)

        np_xysm_1 = numpy.vstack((np_x_1, np_y_d_1, np_y_sd_1, np_y_d_2)).transpose()
        w_s_4 = cwidg_pwd()
        w_s_4.plot_numpy_arrays(np_x_2, np_xysm_1, np_xb_1)

        stack_widg = QtWidgets.QStackedWidget()
        stack_widg.addWidget(w_s_1)
        stack_widg.addWidget(w_s_2)
        stack_widg.addWidget(w_s_3)
        stack_widg.addWidget(w_s_4)
        
        lay_h = QtWidgets.QHBoxLayout()
        _rb_1 = QtWidgets.QRadioButton("sum")
        _rb_1.toggled.connect(lambda: stack_widg.setCurrentIndex(0))
        lay_h.addWidget(_rb_1)
        _rb_1.setChecked(True)
        _rb_2 = QtWidgets.QRadioButton("diff")
        _rb_2.toggled.connect(lambda: stack_widg.setCurrentIndex(1))
        lay_h.addWidget(_rb_2)
        _rb_3 = QtWidgets.QRadioButton("up")
        _rb_3.toggled.connect(lambda: stack_widg.setCurrentIndex(2))
        lay_h.addWidget(_rb_3)
        _rb_4 = QtWidgets.QRadioButton("down")
        _rb_4.toggled.connect(lambda: stack_widg.setCurrentIndex(3))
        lay_h.addWidget(_rb_4)
        lay_h.addStretch(1)

        layout_3.addLayout(lay_h)
        layout_3.addWidget(stack_widg)
    elif (f_meas & f_bkgr):
        np_x_1 = numpy.array(meas.ttheta, dtype = float)
        np_y_u_1 = numpy.array(meas.intensity_up, dtype = float)
        np_y_su_1 = numpy.array(meas.intensity_up_sigma, dtype = float)
        np_y_d_1 = numpy.array(meas.intensity_down, dtype = float)
        np_y_sd_1 = numpy.array(meas.intensity_down_sigma, dtype = float)

        if numpy.all(numpy.isnan(np_y_u_1)):
            np_y_s_1 = numpy.array(meas.intensity, dtype = float)
            np_y_ss_1 = numpy.array(meas.intensity_sigma, dtype = float)
            np_y_m_1 = 0*np_y_s_1
            np_y_sm_1 = np_y_ss_1
        else:
            np_y_s_1 = np_y_u_1 + np_y_d_1
            np_y_ss_1 = numpy.sqrt(numpy.square(np_y_su_1) + numpy.square(np_y_sd_1))
            np_y_m_1 = np_y_u_1 - np_y_d_1
            np_y_sm_1 = np_y_ss_1

        np_x_2 = numpy.array(bkgr.ttheta, dtype = float)
        np_y_b_2 = numpy.array(bkgr.intensity, dtype = float)
        np_xys_1 = numpy.vstack((np_x_1, np_y_s_1, np_y_ss_1)).transpose()
        np_xy_1 = numpy.vstack((np_x_1, np_y_s_1)).transpose()
        np_xb_2 = numpy.vstack((np_x_2, 2*np_y_b_2)).transpose()
        w_s_1 = cwidg_pwd()
        w_s_1.plot_numpy_arrays(np_xy_1, np_xys_1, np_xb_2)

        np_xys_1 = numpy.vstack((np_x_1, np_y_m_1, np_y_sm_1)).transpose()
        np_xy_1 = numpy.vstack((np_x_1, np_y_m_1)).transpose()
        w_s_2 = cwidg_pwd()
        w_s_2.plot_numpy_arrays(np_xy_1, np_xys_1)

        np_xys_1 = numpy.vstack((np_x_1, np_y_u_1, np_y_su_1)).transpose()
        np_xy_1 = numpy.vstack((np_x_1, np_y_u_1)).transpose()
        np_xb_2 = numpy.vstack((np_x_2, np_y_b_2)).transpose()
        w_s_3 = cwidg_pwd()
        w_s_3.plot_numpy_arrays(np_xy_1, np_xys_1, np_xb_2)

        np_xys_1 = numpy.vstack((np_x_1, np_y_d_1, np_y_sd_1)).transpose()
        np_xy_1 = numpy.vstack((np_x_1, np_y_d_1)).transpose()
        w_s_4 = cwidg_pwd()
        w_s_4.plot_numpy_arrays(np_xy_1, np_xys_1, np_xb_2)

        stack_widg = QtWidgets.QStackedWidget()
        stack_widg.addWidget(w_s_1)
        stack_widg.addWidget(w_s_2)
        stack_widg.addWidget(w_s_3)
        stack_widg.addWidget(w_s_4)

        lay_h = QtWidgets.QHBoxLayout()
        _rb_1 = QtWidgets.QRadioButton("sum")
        _rb_1.toggled.connect(lambda: stack_widg.setCurrentIndex(0))
        lay_h.addWidget(_rb_1)
        _rb_1.setChecked(True)
        _rb_2 = QtWidgets.QRadioButton("diff")
        _rb_2.toggled.connect(lambda: stack_widg.setCurrentIndex(1))
        lay_h.addWidget(_rb_2)
        _rb_3 = QtWidgets.QRadioButton("up")
        _rb_3.toggled.connect(lambda: stack_widg.setCurrentIndex(2))
        lay_h.addWidget(_rb_3)
        _rb_4 = QtWidgets.QRadioButton("down")
        _rb_4.toggled.connect(lambda: stack_widg.setCurrentIndex(3))
        lay_h.addWidget(_rb_4)
        lay_h.addStretch(1)

        layout_3.addLayout(lay_h)
        layout_3.addWidget(stack_widg)
    elif f_meas:
        pass

    return 




import numpy
from PyQt5 import QtWidgets, QtGui, QtCore

def widget_for_diffrn_refln(diffrn_refln):
    lay_to_fill = QtWidgets.QHBoxLayout()

    lay_right = QtWidgets.QVBoxLayout()
    size_pol_2 =  QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
    label_out = QtWidgets.QLabel()
    label_out.setSizePolicy(size_pol_2)
    label_out.setFont(QtGui.QFont("Courier", 8, QtGui.QFont.Normal))
    label_out.setTextInteractionFlags(QtCore.Qt.TextSelectableByMouse)
    area = QtWidgets.QScrollArea()
    area.setWidgetResizable(True)
    area.setWidget(label_out)
    lay_right.addWidget(area)

    lay_left = QtWidgets.QVBoxLayout()

    n_row, n_col = len(diffrn_refln.h), 5
    l_name = ["h", "k", "l", "fr", "fr_sigma"]
    _i_row = 0
    l_add = n_row*[None]
    if diffrn_refln.fr_calc is not None:
        if len(diffrn_refln.fr_calc) == n_row:
            l_add  = diffrn_refln.fr_calc
            n_col = 6
            l_name = ["h", "k", "l", "fr", "fr_sigma", "fr_calc"]
    w_t = QtWidgets.QTableWidget(n_row, n_col)
    w_t.setHorizontalHeaderLabels(l_name)
        
    for _h, _k, _l, _fr, _fr_sigma, _fr_calc in zip(diffrn_refln.h, diffrn_refln.k, 
                                               diffrn_refln.l, diffrn_refln.fr, diffrn_refln.fr_sigma, l_add): 
        _w_ti_0 = QtWidgets.QTableWidgetItem()
        _w_ti_0.setText(str(_h))
        _w_ti_1 = QtWidgets.QTableWidgetItem()
        _w_ti_1.setText(str(_k))
        _w_ti_2 = QtWidgets.QTableWidgetItem()
        _w_ti_2.setText(str(_l))
        _w_ti_3 = QtWidgets.QTableWidgetItem()
        _w_ti_3.setText(f"{_fr:.5f}")
        _w_ti_4 = QtWidgets.QTableWidgetItem()
        _w_ti_4.setText(f"{_fr_sigma:.5f}")
        w_t.setItem(_i_row, 0, _w_ti_0)
        w_t.setItem(_i_row, 1, _w_ti_1)
        w_t.setItem(_i_row, 2, _w_ti_2)
        w_t.setItem(_i_row, 3, _w_ti_3)
        w_t.setItem(_i_row, 4, _w_ti_4)
        if _fr_calc is not None:
            _w_ti_5 = QtWidgets.QTableWidgetItem()
            _w_ti_5.setText(f"{_fr_calc:.5f}")
            w_t.setItem(_i_row, 5, _w_ti_5)
        _i_row += 1
    lay_left.addWidget(w_t)


    _b_chi_sq = QtWidgets.QPushButton("calc_chi_sq")
    _b_chi_sq.clicked.connect(lambda : calc_chi_sq(diffrn_refln, label_out))
    lay_left.addWidget(_b_chi_sq)


    _b_info = QtWidgets.QPushButton("info")
    _b_info.clicked.connect(lambda : show_info(diffrn_refln, label_out))
    lay_left.addWidget(_b_info)


    lay_to_fill.addLayout(lay_left)
    lay_to_fill.addLayout(lay_right)
    widg_out = QtWidgets.QWidget()
    widg_out.setLayout(lay_to_fill)
    
    return widg_out

def show_info(diffrn_refln, label):
    label.setText(str(diffrn_refln))

def calc_chi_sq(diffrn_refln, label):
    ls_out = []
    if diffrn_refln.fr_calc is not None:
        _e = numpy.array(diffrn_refln.fr, dtype=float)
        _s = numpy.array(diffrn_refln.fr_sigma, dtype=float)
        _m = numpy.array(diffrn_refln.fr_calc, dtype=float)
        _t = ((_e-_m)/_s)**2
        _n = int((numpy.logical_not(numpy.isnan(_t))).sum())
        chi_sq =  numpy.nansum(_t)
        chi_sq_n = chi_sq/float(_n)
        ls_out.append(f"Number of reflections is {_n:}.")
        ls_out.append(f"Chi_sq per n is {chi_sq_n:.2f}.")
    else:
        ls_out.append(f"Number of reflections is {len(diffrn_refln.fr):}")
        ls_out.append("fr_calc is not defined")
    label.setText("\n".join(ls_out))



import numpy
from PyQt5 import QtWidgets, QtGui, QtCore
from types import FunctionType
from cryspy import Fitable
from cryspy import GlobalConstr
from cryspy import DataConstr, LoopConstr, ItemConstr
from pycifstar import Global, Data, Loop, Item


from cryspy_editor.b_rcif_to_cryspy import L_ITEM_CLASS, L_LOOP_CLASS, L_DATA_CLASS

def temp_func_loop(qt_table, i_row, i_col, obj, l_ii):
    table_item = qt_table.item(i_row, i_col)
    val = str(table_item.text())
    l_attr = obj.ITEM_CLASS.MANDATORY_ATTRIBUTE + obj.ITEM_CLASS.OPTIONAL_ATTRIBUTE
    setattr(obj.item[i_row], l_attr[l_ii[i_col]], val)
    return

def temp_func_loop_help(qt_table, i_row, i_col, obj, l_ii, label):
    table_item = qt_table.item(i_row, i_col)
    val = str(table_item.text())
    l_attr = obj.ITEM_CLASS.MANDATORY_ATTRIBUTE + obj.ITEM_CLASS.OPTIONAL_ATTRIBUTE
    if (label is not None):
        label.setText(get_help_for_attribute(obj.item[i_row], l_attr[l_ii[i_col]]))
    return

def temp_func_item_constr(qt_table, i_row, i_col, obj, l_ii):
    table_item = qt_table.item(i_row, i_col)
    val = str(table_item.text())
    l_attr = obj.MANDATORY_ATTRIBUTE + obj.OPTIONAL_ATTRIBUTE
    if i_col == 1:
        setattr(obj, l_attr[l_ii[i_row]], val)
    return

def temp_func_item(qt_table, i_row, i_col, obj, l_ii):
    table_item = qt_table.item(i_row, i_col)
    val = str(table_item.text())
    if i_col == 1:
        setattr(obj, "value", val)
    return

def temp_func_item_help(qt_table, i_row, i_col, obj, l_ii, label):
    table_item = qt_table.item(i_row, i_col)
    val = str(table_item.text())
    l_attr = obj.MANDATORY_ATTRIBUTE + obj.OPTIONAL_ATTRIBUTE
    if (label is not None):
        label.setText(get_help_for_attribute(obj, l_attr[l_ii[i_row]]))
    return


def make_qtablewidget_for_loop(obj, label=None):

    l_ii, l_attr_sh = [], obj.names
    if l_attr_sh is None:
        return  QtWidgets.QTableWidget(0, 0)

    n_row, n_col = len(obj.values), len(l_attr_sh)
    w_t = QtWidgets.QTableWidget(n_row, n_col)
    l_name = l_attr_sh

    for _i_attr, _attr in enumerate(l_attr_sh):
        _item_header = QtWidgets.QTableWidgetItem(_attr)
        w_t.setHorizontalHeaderItem(_i_attr, _item_header)

    _i_row = 0
    for _i_row, _item in enumerate(obj.values):
        for _i, _attr in enumerate(l_attr_sh):
            _1 = _item[_i]
            _w_ti_0 = QtWidgets.QTableWidgetItem()
            _w_ti_0.setText(str(_1))
            w_t.setItem(_i_row, _i, _w_ti_0)

    #w_t.cellChanged.connect(lambda _1, _2: temp_func_loop(w_t, _1, _2, obj, l_ii))
    return w_t


def make_qtablewidget_for_loop_constr(obj, label=None):
    l_attr = obj.ITEM_CLASS.MANDATORY_ATTRIBUTE + obj.ITEM_CLASS.OPTIONAL_ATTRIBUTE
    l_flag_mandatory = len(obj.ITEM_CLASS.MANDATORY_ATTRIBUTE)*[True]+len(obj.ITEM_CLASS.OPTIONAL_ATTRIBUTE)*[False]
    if len(obj.item)==0:
        n_row, n_col = 0, len(l_attr)
        w_t = QtWidgets.QTableWidget(n_row, n_col)
        for _i_attr, _attr in enumerate(l_attr):
            _item_header = QtWidgets.QTableWidgetItem(_attr)
            if l_flag_mandatory[_i_attr]:
                _item_header.setBackground(QtGui.QColor(255, 255, 240))#Ivory
            else:
                _item_header.setBackground(QtGui.QColor(255, 255, 255))#White
            w_t.setHorizontalHeaderItem(_i_attr, _item_header)
        return w_t

    _item = obj.item[0]
    l_ii, l_attr_sh = [], []
    l_flag_m_sh = []
    for _i, _attr in enumerate(l_attr):
        if getattr(_item, _attr) is not None:
            l_ii.append(_i) 
            l_attr_sh.append(_attr)
            l_flag_m_sh.append(l_flag_mandatory[_i])

    if len(l_attr_sh) == 0:
        l_attr_sh = l_attr
        l_ii = [_i for _i in range(len(l_attr))]

    n_row, n_col = len(obj.item), len(l_attr_sh)
    w_t = QtWidgets.QTableWidget(n_row, n_col)
    l_name = l_attr_sh


    for _i_attr, _attr in enumerate(l_attr_sh):
        _item_header = QtWidgets.QTableWidgetItem(_attr)
        if l_flag_m_sh[_i_attr]:
            _item_header.setBackground(QtGui.QColor(255, 255, 240))#Ivory
        else:
            _item_header.setBackground(QtGui.QColor(255, 255, 255))#White
        w_t.setHorizontalHeaderItem(_i_attr, _item_header)


    _i_row = 0
    for _i_row, _item in enumerate(obj.item):
        for _i, _attr in enumerate(l_attr_sh):
            _1 = getattr(_item, _attr)
            _w_ti_0 = QtWidgets.QTableWidgetItem()
            _w_ti_0.setToolTip(get_help_for_attribute(_item, _attr))
            if isinstance(_1, float):
                _w_ti_0.setText(str(round(_1, 5)))
            elif isinstance(_1, Fitable):
                if _1.refinement:
                    _w_ti_0.setText(str(_1))
                else:
                    _w_ti_0.setText(str(round(float(_1), 5)))
            else:
                _w_ti_0.setText(str(_1))
            w_t.setItem(_i_row, _i, _w_ti_0)

    w_t.cellChanged.connect(lambda _1, _2: temp_func_loop(w_t, _1, _2, obj, l_ii))
    return w_t

def make_qtablewidget_for_item(obj, label=None):
    l_attr_sh = [obj.name]
    l_flag_m_sh = [True]
    n_row, n_col = len(l_attr_sh), 2
    w_t = QtWidgets.QTableWidget(n_row, n_col)

    l_ii = [0]
    l_name = l_attr_sh
    w_t.setHorizontalHeaderLabels(["name", "value"])
    w_t.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Stretch)
    _i_row = 0
    for _i_row, _attr in enumerate(l_attr_sh):
        _w_ti_0 = QtWidgets.QTableWidgetItem()
        _w_ti_0.setText(str(_attr))
        w_t.setItem(_i_row, 0, _w_ti_0)
        _1 = obj.value
        _w_ti_1 = QtWidgets.QTableWidgetItem()
        _w_ti_1.setText(str(_1))
        if l_flag_m_sh[_i_row]:
            _w_ti_0.setBackground(QtGui.QColor(255, 255, 240))#Ivory
            _w_ti_1.setBackground(QtGui.QColor(255, 255, 240))
        else:
            _w_ti_0.setBackground(QtGui.QColor(255, 255, 255))#White
            _w_ti_1.setBackground(QtGui.QColor(255, 255, 255))

        w_t.setItem(_i_row, 1, _w_ti_1)
    w_t.cellChanged.connect(lambda _1, _2: temp_func_item(w_t, _1, _2, obj, l_ii))
    w_t.setVerticalHeaderLabels(n_row*[""])
    return w_t


def make_qtablewidget_for_item_constr(obj, label=None):
    l_attr = obj.MANDATORY_ATTRIBUTE + obj.OPTIONAL_ATTRIBUTE
    l_flag_mandatory = len(obj.MANDATORY_ATTRIBUTE)*[True]+len(obj.OPTIONAL_ATTRIBUTE)*[False]
    l_ii, l_attr_sh = [], []
    l_flag_m_sh = []
    l_ii_end, l_attr_sh_end = [], []
    l_flag_m_sh_end = []
    for _i, _attr in enumerate(l_attr):
        if getattr(obj, _attr) is not None:
            l_ii.append(_i) 
            l_attr_sh.append(_attr)
            l_flag_m_sh.append(l_flag_mandatory[_i])
        else:
            l_ii_end.append(_i)
            l_attr_sh_end.append(_attr)
            l_flag_m_sh_end.append(l_flag_mandatory[_i])

    l_ii.extend(l_ii_end)
    l_attr_sh.extend(l_attr_sh_end)
    l_flag_m_sh.extend(l_flag_m_sh_end)

    n_row, n_col = len(l_attr_sh), 2
    w_t = QtWidgets.QTableWidget(n_row, n_col)
    l_name = l_attr_sh
    w_t.setHorizontalHeaderLabels(["name", "value"])
    w_t.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Stretch)
    _i_row = 0
    for _i_row, _attr in enumerate(l_attr_sh):
        _w_ti_0 = QtWidgets.QTableWidgetItem()
        _w_ti_0.setText(str(_attr))
        w_t.setItem(_i_row, 0, _w_ti_0)
        _1 = getattr(obj, _attr)
        _w_ti_0.setToolTip(get_help_for_attribute(obj, _attr))
        _w_ti_1 = QtWidgets.QTableWidgetItem()
        _w_ti_1.setToolTip(_1.__doc__)
        _w_ti_1.setText(str(_1))
        _w_ti_1.setToolTip(get_help_for_attribute(obj, _attr))
        if l_flag_m_sh[_i_row]:
            _w_ti_0.setBackground(QtGui.QColor(255, 255, 240))#Ivory
            _w_ti_1.setBackground(QtGui.QColor(255, 255, 240))
        else:
            _w_ti_0.setBackground(QtGui.QColor(255, 255, 255))#White
            _w_ti_1.setBackground(QtGui.QColor(255, 255, 255))

        w_t.setItem(_i_row, 1, _w_ti_1)
    w_t.cellChanged.connect(lambda _1, _2: temp_func_item_constr(w_t, _1, _2, obj, l_ii))
    w_t.setVerticalHeaderLabels(n_row*[""])
    return w_t


def make_qtablewidget_for_data_constr(obj, label, thread):
    l_data_obj = obj.mandatory_objs + obj.optional_objs
    l_ii, l_obj_sh = [], []

    l_obj_sh = l_data_obj
    l_ii = [_i for _i in range(len(l_data_obj))]

    n_row, n_col = len(l_obj_sh)+1, 1
    w_t = QtWidgets.QTableWidget(n_row, n_col)
    
    w_t.setHorizontalHeaderLabels(["name"])
    w_t.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Stretch)
    for _i_row, _obj in enumerate(l_obj_sh):
        _w_ti_0 = QtWidgets.QTableWidgetItem()
        s_val = str(type(_obj).__name__)
        if isinstance(_obj, GlobalConstr): 
            s_val += f": {_obj.global_name:}"
        elif isinstance(_obj, DataConstr): 
            s_val += f": {_obj.data_name:}"
        elif isinstance(_obj, LoopConstr): 
            s_val += f": {_obj.loop_name:}"
        _w_ti_0.setText(s_val)
        _w_ti_0.setToolTip(_obj.__doc__)
        _w_ti_0.setCheckState(0)
        w_t.setItem(_i_row, 0, _w_ti_0)
    pb_del = QtWidgets.QPushButton("Delete")
    w_t.setCellWidget(n_row-1, 0, pb_del)
    
    w_t.setVerticalHeaderLabels(n_row*[""])
    pb_del.clicked.connect(lambda : del_obj_by_table(w_t, obj, thread))
    return w_t

def del_obj_by_table(table_widget, obj, thread):
    n_mandatory = len(obj.mandatory_objs)
    n_optional = len(obj.optional_objs)
    n_row = table_widget.rowCount()
    l_del = []
    l_ind = []
    for i_row in range(n_row-1):
        val = 0
        flag=False
        _table_item =  table_widget.item(i_row, 0)
        if _table_item is not None:
            l_ind.append(i_row)
            val = _table_item.checkState()
        if val == 2: flag = True
        l_del.append(flag)
    for _1, _2 in zip(range((n_row-2), -1, -1), reversed(l_del)):
        if _2:
            table_widget.takeItem(_1, 0)
            _ind = l_ind.index(_1)
            if _ind < n_mandatory:
                n_mandatory -= 1
                obj.mandatory_objs.pop(_ind)
            else:
                n_optional -= 1
                obj.optional_objs.pop(_ind-n_mandatory)
    thread.signal_refresh.emit()   

def show_info(obj, label):
    if obj.is_defined:
        obj.form_object
    label.setText(str(obj))

def show_variables(obj, label):
    if obj.is_defined:
        obj.form_object
    ls_str= [f"Variables:\n"]
    l_variable = obj.get_variables()
    for _ind, _variable in enumerate(l_variable):
        ls_str.append(f"({_ind+1:}) {str(_variable):}")
    label.setText("\n".join(ls_str))

def show_help(obj, label):
    ls_out = []
    ls_out.append(str(obj.__doc__))
    label.setText("\n".join(ls_out))

def get_help_for_attribute(obj, attribute):
    ls_out = []
    obj_class = type(obj)
    try:
        func = getattr(obj_class, attribute)
        if isinstance(func, FunctionType):
            l_param = [ _ for _ in func.__code__.co_varnames[:func.__code__.co_argcount] if _ != "self"]
            if len(l_param) > 0:
                ls_out.append("INPUT PARAMETERS: \n"+ "\n".join(l_param))
            else:
                ls_out.append("NO INPUT PARAMETERS.")
            if func.__defaults__ is not None:
                ls_out.append("\nDEFAULT PARAMETERS: \n" + "\n".join([str(_) for _ in func.__defaults__]))
            #print("func.__kwdefaults__: ", func.__kwdefaults__)
            #print("func.__code__.co_kwonlyargcount: ", func.__code__.co_kwonlyargcount)
        ls_out.append(70*"*")
        val = func.__doc__
        if val is not None:
            ls_out.append(val)
        else:
            ls_out.append("Documentation is not written")
    except:
        ls_out.append("It's not found.")
    return "\n".join(ls_out)

def get_method_list(obj):
    _cls = type(obj)
    l_method = [_1 for _1, _2 in _cls.__dict__.items() if ((type(_2) == FunctionType) & (not(_1.startswith("_"))))]
    l_method.sort()
    return l_method

def get_layout_method_help(obj, label_out, thread):
    _lay = QtWidgets.QVBoxLayout()
    
    l_method = get_method_list(obj)

    if len(l_method) != 0:
        obj_class = type(obj)
        _label = QtWidgets.QLabel("Methods (double click to do): ")
        _lay.addWidget(_label)

        _list_widget = QtWidgets.QListWidget()
        qfont = QtGui.QFont()
        qfont.setBold(True)
        for _method in l_method:
            func = getattr(obj_class, _method)
            l_param = [_ for _ in func.__code__.co_varnames[:func.__code__.co_argcount] if _ != "self"]
            s_par = ""
            if len(l_param) > 0:
                s_par = ", ".join(l_param)
            s_val = f"{_method:}({s_par:})"

            _list_widget_item = QtWidgets.QListWidgetItem(s_val)
            _list_widget_item.setFont(qfont)
            _list_widget_item.setToolTip(get_help_for_attribute(obj, _method))
            _list_widget.addItem(_list_widget_item)
        _list_widget.setSortingEnabled(True)
        _lay.addWidget(_list_widget)
        table_widget_1 = QtWidgets.QTableWidget()
        table_widget_1.setHorizontalHeaderLabels(["name", "value"])
        table_widget_1.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Stretch)
        table_widget_1.hide()
        _lay.addWidget(table_widget_1)

        _list_widget.doubleClicked.connect(lambda : do_obj_func(obj, _list_widget.currentItem().text().split("(")[0], table_widget_1, label_out, thread))
    return _lay

def do_obj_func(obj, func_name, table_widget, label, thread):
    ls_out = []
    func = getattr(obj, func_name)
    if (func.__code__.co_argcount == 1):
        #res = func()
        ls_out.append("Calculations are started")
        thread.function = func
        thread.arguments = ()
        thread.start()
        table_widget.hide()
    else: 
        flag = False
        if flag:
            ls_out.append("Function is performed with default parameters:\n")
            ls_out.append(str(res))
            table_widget.hide()
        else:
            table_widget.show()
            n_row_need = (func.__code__.co_argcount-1) 
            n_row = table_widget.rowCount()
            flag_read = n_row_need==n_row

            l_var_names = func.__code__.co_varnames[1:(n_row_need+1)]

            if flag_read:
                flag_read = all([str(table_widget.item(_i_row, 0).text()) == _2 for _i_row, _2 in zip(range(n_row), l_var_names)])
                flag_read_2 = all([str(table_widget.item(_i_row, 1).text()) != "" for _i_row in range(n_row)])
                flag_read = flag_read & flag_read_2

            if func.__defaults__ is None:
                n_defaults = 0
            else:
                n_defaults = len(func.__defaults__)
            n_var_names = len(l_var_names)
            
            if not(flag_read):
                table_widget.setColumnCount(2)
                table_widget.setRowCount(n_row_need)
                for _i, _name in enumerate(l_var_names):
                    t_w_i_1 = table_widget.item(_i, 0)
                    if t_w_i_1 is None:
                        t_w_i_1 = QtWidgets.QTableWidgetItem()
                        table_widget.setItem(_i, 0, t_w_i_1)
                    t_w_i_1.setText(_name)
                    t_w_i_2 = table_widget.item(_i, 1)
                    if t_w_i_2 is None:
                        t_w_i_2 = QtWidgets.QTableWidgetItem()
                        table_widget.setItem(_i, 1, t_w_i_2)
                    i_default = _i-(n_var_names-n_defaults)
                    if i_default >= 0:
                        t_w_i_2.setText(str(func.__defaults__[i_default]))

                ls_out.append("Introduce the input parameters in the table.")
            else:
                l_var_in = []
                for _i, _name in enumerate(l_var_names):
                    s_val = table_widget.item(_i, 1).text()
                    try:
                        e_val = eval(s_val)
                        if isinstance(e_val, (str, list, tuple, float, int)):
                            val_in = numpy.array(e_val, dtype=float)
                        else:
                            val_in = e_val
                    except:
                        val_in = s_val.strip()
                    if s_val == "None":
                        val_in = None
                    l_var_in.append(val_in)
                t_var_in = tuple(l_var_in)

                ls_out.append("Calculations are started")
                thread.function = func
                thread.arguments = t_var_in
                thread.start()

    label.setText("\n".join(ls_out))

def get_layout_internal_attribute(obj, label_out):
    _lay = QtWidgets.QVBoxLayout()
    
    l_attr_name = obj.INTERNAL_ATTRIBUTE
    if l_attr_name is None:
        return _lay
    l_attr_name = list(l_attr_name)
    l_attr_name.sort()
    if len(l_attr_name) != 0:
        _label = QtWidgets.QLabel("Internal attributes: ")
        _lay.addWidget(_label)

        _list_widget = QtWidgets.QListWidget()
        for _attr_name in l_attr_name:
            val = str(getattr(obj, _attr_name))
            if isinstance(val, float):
                s_val = f"{_attr_name:}: {round(val, 5):}"
            elif isinstance(val, Fitable):
                if val.refinement:
                    s_val = f"{_attr_name:}: {str(val):}"
                else:
                    s_val = f"{_attr_name:}: {round(float(val), 5):}"
            elif len(str(val).split("\n")) > 1:
                s_val = f"{_attr_name:}: ..."
            elif len(str(val)) > 10:
                try:
                    s_val = f"{_attr_name:}: {round(float(val), 5):}"
                except:
                    s_val = f"{_attr_name:}: {str(val)[:7]:}..."
            else:
                s_val = f"{_attr_name:}: {str(val):}"

            _list_widget_item = QtWidgets.QListWidgetItem(s_val)
            _list_widget_item.setToolTip(get_attribute_with_help(obj, _attr_name))
            _list_widget.addItem(_list_widget_item)
        _list_widget.itemDoubleClicked.connect(lambda _1: label_out.setText(get_attribute_with_help(obj, _1.text().split(":")[0])))
        _lay.addWidget(_list_widget)
    
    return _lay

def get_attribute_with_help(obj, attribute):
    ls_out = []
    ls_out.append(f"{attribute:}:\n")
    ls_out.append(str(getattr(obj, attribute)))
    ls_out.append("\n"+50*"*"+"\n")
    ls_out.append(get_help_for_attribute(obj, attribute))
    return "\n".join(ls_out)


def del_layout(layout):
    """
delete all elements from layouts
    """
    for i in reversed(range(layout.count())):
        if layout.itemAt(i).widget() != None:
           layout.itemAt(i).widget().setParent(None)
        elif layout.itemAt(i).layout() != None:
            del_layout(layout.itemAt(i).layout())
            layout.itemAt(i).layout().setParent(None)
        else:
            layout.removeItem(layout.itemAt(i))
    return


def show_widget(widget, state):
    if int(state) == 0:
        widget.hide()
    else:
        widget.show()


def add_mandatory_optional_obj(obj, label, thread):
    """
layout to add new attributes for DataConstr or GlobalConstr
    """
    layout = QtWidgets.QVBoxLayout()
    layout.addWidget(QtWidgets.QLabel("Add attribute:"))


    if type(obj) == GlobalConstr:
        l_cls = L_ITEM_CLASS + L_LOOP_CLASS + L_DATA_CLASS
        l_flag_mandatory = len(l_cls)*[False]
    elif type(obj) == DataConstr:
        l_cls = L_ITEM_CLASS + L_LOOP_CLASS
        l_flag_mandatory = len(l_cls) * [False]
    else:
        l_cls = obj.mandatory_classes+obj.optional_classes
        l_flag_mandatory = len(obj.mandatory_classes) * [True] + len(obj.optional_classes) * [False]

    l_cls_name_module_bases = [(_, _.__name__, _.__module__, _.__bases__[0].__name__, str(int(not(_f)))) for _, _f in zip(l_cls, l_flag_mandatory)]
    l_cls_name_module_bases.sort(key=lambda x: x[4]+x[3]+x[2]+x[1])
    l_cls = [_[0] for _ in l_cls_name_module_bases]
    n_row, n_col = len(l_cls)+1, 3
    table_widget_1 = QtWidgets.QTableWidget(n_row, n_col)
    table_widget_1.setHorizontalHeaderLabels(["class_name", "type", "module"])


    header = table_widget_1.horizontalHeader()
    header.setSectionResizeMode(0, QtWidgets.QHeaderView.Stretch)
    header.setSectionResizeMode(1, QtWidgets.QHeaderView.ResizeToContents)
    header.setSectionResizeMode(2, QtWidgets.QHeaderView.ResizeToContents)


    for i_row, _cls_name_module_bases in enumerate(l_cls_name_module_bases):
        
        s_name = _cls_name_module_bases[1]
        s_module = ".".join(_cls_name_module_bases[2].split(".")[:-1])
        s_bases = _cls_name_module_bases[3]
        s_flag_m = _cls_name_module_bases[4]

        if s_bases == "GlobalConstr":
            s_bases = "global block "
        elif s_bases == "DataConstr":
            s_bases = "data block "
        elif s_bases == "LoopConstr":
            s_bases = "loop "
        elif s_bases == "ItemConstr":
            s_bases = "items "
        s_doc = _cls_name_module_bases[0].__doc__

        table_widget_1_item = QtWidgets.QTableWidgetItem(s_name)
        table_widget_1_item.setCheckState(0)
        table_widget_1_item.setToolTip(s_doc)
        table_widget_1.setItem(i_row, 0, table_widget_1_item)

        table_widget_2_item = QtWidgets.QTableWidgetItem(s_bases)
        table_widget_2_item.setToolTip(s_doc)
        table_widget_1.setItem(i_row, 1, table_widget_2_item)

        table_widget_3_item = QtWidgets.QTableWidgetItem(s_module)
        table_widget_3_item.setToolTip(s_doc)
        table_widget_1.setItem(i_row, 2, table_widget_3_item)

        if s_flag_m=="0":
            table_widget_1_item.setBackground(QtGui.QColor(255, 255, 240))#Ivory
            table_widget_2_item.setBackground(QtGui.QColor(255, 255, 240))
            table_widget_3_item.setBackground(QtGui.QColor(255, 255, 240))
        else:
            table_widget_1_item.setBackground(QtGui.QColor(255, 255, 255))#White
            table_widget_2_item.setBackground(QtGui.QColor(255, 255, 255))
            table_widget_3_item.setBackground(QtGui.QColor(255, 255, 255))


    pb_add = QtWidgets.QPushButton("Add objects")
    table_widget_1.setCellWidget(n_row-1, 0, pb_add)
    table_widget_1.setVerticalHeaderLabels(n_row*[""])
    pb_add.clicked.connect(lambda : add_obj_by_table(table_widget_1, l_cls, obj, thread))

    layout.addWidget(table_widget_1)

    return layout


def add_obj_by_table(table_widget, l_cls, obj, thread):
    if type(obj) == GlobalConstr:
        flag_constrained = True
    elif type(obj) == DataConstr:
        flag_constrained = True
    else:
        flag_constrained = False
    
    n_mandatory = len(obj.mandatory_classes)
    n_optional = len(obj.optional_classes)
    n_row = table_widget.rowCount()
    l_del = []
    l_ind = []
    l_name = []
    for i_row in range(n_row-1):
        val = 0
        flag=False
        _table_item =  table_widget.item(i_row, 0)
        if _table_item is not None:
            l_ind.append(i_row)
            val = _table_item.checkState()
        
        _table_item_2 =  table_widget.item(i_row, 1)
        if val == 2: flag = True
        l_del.append(flag)
        s_name = ""
        l_name.append(s_name)
    for _1, _2, _name in zip(range((n_row-2), -1, -1), reversed(l_del), reversed(l_name)):
        if _2:
            _ind = l_ind.index(_1)
            _cls = l_cls[_ind]
            _item = _cls()
            if isinstance(_item, GlobalConstr):
                _item.global_name=_name
            elif isinstance(_item, DataConstr):
                _item.data_name=_name
            elif isinstance(_item, LoopConstr):
                _item.loop_name=_name

            if _item is not None:
                l_add_mandatory_obj = []
                l_add_optional_obj = []
                flag = False
                if _ind < n_mandatory:
                    for _item_in in obj.mandatory_objs:
                        if type(_item_in) == type(_item):
                            if isinstance(_item_in, GlobalConstr):
                                if _item_in.global_name == _name:
                                    obj.mandatory_objs.remove(_item_in)
                            elif isinstance(_item_in, DataConstr):
                                if _item_in.data_name == _name:
                                    obj.mandatory_objs.remove(_item_in)
                            elif isinstance(_item_in, LoopConstr):
                                if _item_in.loop_name == _name:
                                    obj.mandatory_objs.remove(_item_in)
                            elif isinstance(_item_in, ItemConstr):
                                obj.mandatory_objs.remove(_item_in)
                    l_add_mandatory_obj.append(_item)
                else:
                    for _item_in in obj.optional_objs:
                        if type(_item_in) == type(_item):
                            if isinstance(_item_in, GlobalConstr):
                                if _item_in.global_name == _name:
                                    obj.optional_objs.remove(_item_in)
                            elif isinstance(_item_in, DataConstr):
                                if _item_in.data_name == _name:
                                    obj.optional_objs.remove(_item_in)
                            elif isinstance(_item_in, LoopConstr):
                                if _item_in.loop_name == _name:
                                    obj.optional_objs.remove(_item_in)
                            elif isinstance(_item_in, ItemConstr):
                                obj.optional_objs.remove(_item_in)
                    l_add_optional_obj.append(_item)

                obj.mandatory_objs.extend(l_add_mandatory_obj)
                obj.optional_objs.extend(l_add_optional_obj)

                if flag_constrained:
                    for _obj in l_add_optional_obj:
                        if not(isinstance(_obj, tuple(obj.optional_classes))):
                            obj.optional_classes.append(type(_obj))
    thread.signal_refresh.emit()

def get_layout_rciftab_obj(obj, thread):
    layout = QtWidgets.QVBoxLayout()

    _text_edit = QtWidgets.QTextEdit()
    if (isinstance(obj, Global) | isinstance(obj, Data) | isinstance(obj, Loop) | isinstance(obj, Item)):
        _text_edit.setText(str(obj))
    else:
        _text_edit.setText(obj.to_cif(flag=True))
    layout.addWidget(_text_edit)

    lay_button = QtWidgets.QHBoxLayout()
    lay_button.addStretch(1)
    button_to_obj = QtWidgets.QPushButton("add to object")
    lay_button.addWidget(button_to_obj)
    layout.addLayout(lay_button)
    button_to_obj.clicked.connect(lambda : add_rcif_to_obj(obj, _text_edit.toPlainText(), thread))
    return layout

def add_rcif_to_obj(obj, text, thread):
    obj_cls = type(obj)
    obj_new = None
    if (isinstance(obj, Global) | isinstance(obj, Data) | isinstance(obj, Loop) | isinstance(obj, Item)):
        l_obj_new = obj_cls() 
        l_obj_new.take_from_string(text)
    else:
        l_obj_new = obj_cls.from_cif(text)
    if l_obj_new is not None:
        if isinstance(l_obj_new, list):
            if len(l_obj_new) != 0:
                obj_new = l_obj_new[0]
        else:
            obj_new = l_obj_new 

    if obj_new is None:
        return
    if isinstance(obj, LoopConstr):
        obj.item = obj_new.item
        obj.form_object
    elif isinstance(obj, DataConstr):
        pass
    elif isinstance(obj, GlobalConstr):
        pass
    elif isinstance(obj, ItemConstr):
        obj.clean_attribute
        l_attr = obj_new.mandatory_attribute + obj.optional_attribute
        for _attr in l_attr:
            if obj_new.is_defined_attribute(_attr):
                setattr(obj, f"__{_attr:}", getattr(obj_new, f"__{_attr:}"))
        if obj.is_defined:
            obj.form_object
    elif isinstance(obj, Item):
        obj.name = obj_new.name
        obj.value = obj_new.value
    elif isinstance(obj, Loop):
        pass
    elif isinstance(obj, Data):
        pass
    elif isinstance(obj, Global):
        pass
    thread.signal_refresh.emit()
    return